"""Wrapper of the `enorm` function."""

from numba import extending, njit
from numpy import floating, int32
from numpy.typing import NDArray

from ._cminpack import Cminpack

# --------------------------------------- enorm -------------------------------------- #


def _enorm(n, x):
    raise NotImplementedError


@extending.overload(_enorm)
def _enorm_overload(n, x):
    enorm_external = Cminpack.enorm(x.dtype)

    def impl(n, x):
        if x.ndim != 1:
            msg = "x must be a 1D array"
            raise ValueError(msg)
        return enorm_external(n, x.ctypes)

    return impl


@njit
def enorm_(n: int32, x: NDArray[floating]) -> float:
    """Euclidean norm of the n-vector x.

    Parameters
    ----------
    n : int
        the length of x
    x : NDArray[floating]
        the vector

    Returns
    -------
    float
        the Euclidean norm of x

    """
    return _enorm(n, x)


@njit
def enorm(x: NDArray[floating]) -> float:
    """Euclidean norm of the n-vector x.

    Parameters
    ----------
    x : NDArray[floating]
        the vector

    Returns
    -------
    float
        the Euclidean norm of x

    """
    return _enorm(x.size, x)



"""Wrappers of the `lmder` and `lmder1` functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from numba import extending, njit, types
from numpy import empty, floating, int32, ones

from ._cminpack import Cminpack
from .utils import _check_dtype, ptr_from_val, ptr_int32, val_from_ptr

if TYPE_CHECKING:
    from numpy import int64
    from numpy.typing import NDArray

# -------------------------------------- lmder1 -------------------------------------- #


def _lmder1(fcn, m, n, x, fvec, fjac, ldfjac, tol, ipvt, wa, lwa, udata):
    raise NotImplementedError


@extending.overload(_lmder1)
def _lmder1_overload(
    fcn,
    m,
    n,
    x,
    fvec,
    fjac,
    ldfjac,
    tol,
    ipvt,
    wa,
    lwa,
    udata,
):
    _check_dtype((fvec, fjac, wa), x.dtype)
    lmder1_external = Cminpack.lmder1(x.dtype)

    @extending.register_jitable
    def impl(fcn, m, n, x, fvec, fjac, ldfjac, tol, ipvt, wa, lwa, udata):
        info = lmder1_external(
            fcn,
            udata,
            m,
            n,
            x.ctypes,
            fvec.ctypes,
            fjac.ctypes,
            ldfjac,
            tol,
            ipvt.ctypes,
            wa.ctypes,
            lwa,
        )
        return x, fvec, fjac, ipvt, info

    if isinstance(udata, types.Array):
        return lambda fcn, m, n, x, fvec, fjac, ldfjac, tol, ipvt, wa, lwa, udata: impl(
            fcn,
            m,
            n,
            x,
            fvec,
            fjac,
            ldfjac,
            tol,
            ipvt,
            wa,
            lwa,
            udata.ctypes,
        )
    if udata is not types.none:
        return impl
    return lambda fcn, m, n, x, fvec, fjac, ldfjac, tol, ipvt, wa, lwa, udata: impl(
        fcn,
        m,
        n,
        x,
        fvec,
        fjac,
        ldfjac,
        tol,
        ipvt,
        wa,
        lwa,
        0,
    )


@njit
def lmder1_(
    fcn: int64,
    m: int32,
    n: int32,
    x: NDArray[floating],
    fvec: NDArray[floating],
    fjac: NDArray[floating],
    ldfjac: int32,
    tol: floating,
    ipvt: NDArray[int32],
    wa: NDArray[floating],
    lwa: int32,
    udata: NDArray | None = None,
) -> tuple[
    NDArray[floating],
    NDArray[floating],
    NDArray[floating],
    NDArray[int32],
    int,
]:
    # TODO(nin17): docstring
    """.

    Parameters
    ----------
    fcn : int64
        _description_
    m : int32
        _description_
    n : int32
        _description_
    x : NDArray[floating]
        _description_
    fvec : NDArray[floating]
        _description_
    fjac : NDArray[floating]
        _description_
    ldfjac : int32
        _description_
    tol : floating
        _description_
    ipvt : NDArray[int32]
        _description_
    wa : NDArray[floating]
        _description_
    lwa : int32
        _description_
    udata : NDArray | None, optional
        _description_, by default None

    Returns
    -------
    tuple[ NDArray[floating], NDArray[floating], NDArray[floating], NDArray[int32], int]
        _description_

    """
    return _lmder1(fcn, m, n, x, fvec, fjac, ldfjac, tol, ipvt, wa, lwa, udata)


@njit
def lmder1(
    fcn: int64,
    m: int32,
    x: NDArray[floating],
    tol: floating | None = None,
    udata: NDArray | None = None,
) -> tuple[
    NDArray[floating],
    NDArray[floating],
    NDArray[floating],
    NDArray[int32],
    int,
]:
    # TODO(nin17): docstring
    """.

    Parameters
    ----------
    fcn : int64
        _description_
    m : int32
        _description_
    x : NDArray[floating]
        _description_
    tol : floating | None, optional
        _description_, by default None
    udata : NDArray | None, optional
        _description_, by default None

    Returns
    -------
    tuple[ NDArray[floating], NDArray[floating], NDArray[floating], NDArray[int32], int]
        _description_

    """
    tol = tol or 1.49012e-8
    n = int32(x.size)
    lwa = int32(5 * n + m)
    fvec = empty(m, dtype=x.dtype)
    fjac = empty((m, n), dtype=x.dtype)
    wa = empty(lwa, dtype=x.dtype)
    ipvt = empty(n, dtype=int32)
    return _lmder1(
        fcn,
        m,
        n,
        x.copy(),
        fvec,
        fjac,
        m,
        tol,
        ipvt,
        wa,
        lwa,
        udata,
    )


# --------------------------------------- lmder -------------------------------------- #


def _lmder(
    fcn,
    m,
    n,
    x,
    fvec,
    fjac,
    ldfjac,
    ftol,
    xtol,
    gtol,
    maxfev,
    diag,
    mode,
    factor,
    nprint,
    nfev,
    njev,
    ipvt,
    qtf,
    wa1,
    wa2,
    wa3,
    wa4,
    udata,
):
    raise NotImplementedError


@extending.overload(_lmder)
def _lmder_overload(
    fcn,
    m,
    n,
    x,
    fvec,
    fjac,
    ldfjac,
    ftol,
    xtol,
    gtol,
    maxfev,
    diag,
    mode,
    factor,
    nprint,
    nfev,
    njev,
    ipvt,
    qtf,
    wa1,
    wa2,
    wa3,
    wa4,
    udata,
):
    _check_dtype((fvec, fjac, qtf, wa1, wa2, wa3, wa4), x.dtype)
    lmder_external = Cminpack.lmder(x.dtype)

    @extending.register_jitable
    def impl(
        fcn,
        m,
        n,
        x,
        fvec,
        fjac,
        ldfjac,
        ftol,
        xtol,
        gtol,
        maxfev,
        diag,
        mode,
        factor,
        nprint,
        nfev,
        njev,
        ipvt,
        qtf,
        wa1,
        wa2,
        wa3,
        wa4,
        udata,
    ):
        info = lmder_external(
            fcn,
            udata,
            m,
            n,
            x.ctypes,
            fvec.ctypes,
            fjac.ctypes,
            ldfjac,
            ftol,
            xtol,
            gtol,
            maxfev,
            diag.ctypes,
            mode,
            factor,
            nprint,
            nfev,
            njev,
            ipvt.ctypes,
            qtf.ctypes,
            wa1.ctypes,
            wa2.ctypes,
            wa3.ctypes,
            wa4.ctypes,
        )
        _nfev = val_from_ptr(nfev)
        _njev = val_from_ptr(njev)
        return x, fvec, fjac, ipvt, qtf, _nfev, _njev, info

    if isinstance(udata, types.Array):
        return (
            lambda fcn,
            m,
            n,
            x,
            fvec,
            fjac,
            ldfjac,
            ftol,
            xtol,
            gtol,
            maxfev,
            diag,
            mode,
            factor,
            nprint,
            nfev,
            njev,
            ipvt,
            qtf,
            wa1,
            wa2,
            wa3,
            wa4,
            udata: impl(
                fcn,
                m,
                n,
                x,
                fvec,
                fjac,
                ldfjac,
                ftol,
                xtol,
                gtol,
                maxfev,
                diag,
                mode,
                factor,
                nprint,
                nfev,
                njev,
                ipvt,
                qtf,
                wa1,
                wa2,
                wa3,
                wa4,
                udata.ctypes,
            )
        )
    if udata is not types.none:
        return impl
    return (
        lambda fcn,
        m,
        n,
        x,
        fvec,
        fjac,
        ldfjac,
        ftol,
        xtol,
        gtol,
        maxfev,
        diag,
        mode,
        factor,
        nprint,
        nfev,
        njev,
        ipvt,
        qtf,
        wa1,
        wa2,
        wa3,
        wa4,
        udata: impl(
            fcn,
            m,
            n,
            x,
            fvec,
            fjac,
            ldfjac,
            ftol,
            xtol,
            gtol,
            maxfev,
            diag,
            mode,
            factor,
            nprint,
            nfev,
            njev,
            ipvt,
            qtf,
            wa1,
            wa2,
            wa3,
            wa4,
            0,
        )
    )


@njit
def lmder_(
    fcn: int64,
    m: int32,
    n: int32,
    x: NDArray[floating],
    fvec: NDArray[floating],
    fjac: NDArray[floating],
    ldfjac: int32,
    ftol: floating,
    xtol: floating,
    gtol: floating,
    maxfev: int32,
    diag: NDArray[floating],
    mode: int32,
    factor: floating,
    nprint: int32,
    nfev: ptr_int32,
    njev: ptr_int32,
    ipvt: NDArray[int32],
    qtf: NDArray[floating],
    wa1: NDArray[floating],
    wa2: NDArray[floating],
    wa3: NDArray[floating],
    wa4: NDArray[floating],
    udata: NDArray | None = None,
) -> tuple[
    NDArray[floating],
    NDArray[floating],
    NDArray[floating],
    NDArray[int32],
    NDArray[floating],
    int,
    int,
    int,
]:
    # TODO(nin17): docstring
    """.

    Parameters
    ----------
    fcn : int64
        _description_
    m : int32
        _description_
    n : int32
        _description_
    x : NDArray[floating]
        _description_
    fvec : NDArray[floating]
        _description_
    fjac : NDArray[floating]
        _description_
    ldfjac : int32
        _description_
    ftol : floating
        _description_
    xtol : floating
        _description_
    gtol : floating
        _description_
    maxfev : int32
        _description_
    diag : NDArray[floating]
        _description_
    mode : int32
        _description_
    factor : floating
        _description_
    nprint : int32
        _description_
    nfev : ptr_int32
        _description_
    njev : ptr_int32
        _description_
    ipvt : NDArray[int32]
        _description_
    qtf : NDArray[floating]
        _description_
    wa1 : NDArray[floating]
        _description_
    wa2 : NDArray[floating]
        _description_
    wa3 : NDArray[floating]
        _description_
    wa4 : NDArray[floating]
        _description_
    udata : NDArray | None, optional
        _description_, by default None

    Returns
    -------
    tuple[ NDArray[floating], NDArray[floating], NDArray[floating], NDArray[int32],
           NDArray[floating], int, int, int]
        _description_

    """
    return _lmder(
        fcn,
        m,
        n,
        x,
        fvec,
        fjac,
        ldfjac,
        ftol,
        xtol,
        gtol,
        maxfev,
        diag,
        mode,
        factor,
        nprint,
        nfev,
        njev,
        ipvt,
        qtf,
        wa1,
        wa2,
        wa3,
        wa4,
        udata,
    )


@njit
def lmder(
    fcn: int64,
    m: int32,
    x: NDArray[floating],
    ftol: floating | None = None,
    xtol: floating | None = None,
    gtol: floating | None = None,
    maxfev: int32 | None = None,
    diag: NDArray[floating] | None = None,
    mode: int32 | None = None,
    factor: floating | None = None,
    nprint: int32 | None = None,
    udata: NDArray | None = None,
) -> tuple[
    NDArray[floating],
    NDArray[floating],
    NDArray[floating],
    NDArray[int32],
    NDArray[floating],
    int,
    int,
    int,
]:
    # TODO(nin17): docstring
    """.

    Parameters
    ----------
    fcn : int64
        _description_
    m : int32
        _description_
    x : NDArray[floating]
        _description_
    ftol : floating | None, optional
        _description_, by default None
    xtol : floating | None, optional
        _description_, by default None
    gtol : floating | None, optional
        _description_, by default None
    maxfev : int32 | None, optional
        _description_, by default None
    diag : NDArray[floating] | None, optional
        _description_, by default None
    mode : int32 | None, optional
        _description_, by default None
    factor : floating | None, optional
        _description_, by default None
    nprint : int32 | None, optional
        _description_, by default None
    udata : NDArray | None, optional
        _description_, by default None

    Returns
    -------
    tuple[ NDArray[floating], NDArray[floating], NDArray[floating], NDArray[int32],
           NDArray[floating], int, int, int]
        _description_

    """
    n = int32(x.size)
    fvec = empty(m, dtype=x.dtype)
    fjac = empty((m, n), dtype=x.dtype)
    ldfjac = m
    nfevptr = ptr_from_val(int32(0))
    njevptr = ptr_from_val(int32(0))
    ipvt = empty(n, dtype=int32)
    qtf = empty(n, dtype=x.dtype)
    wa = empty(3 * n + m, dtype=x.dtype)
    wa1 = wa[:n]
    wa2 = wa[n : 2 * n]
    wa3 = wa[2 * n : 3 * n]
    wa4 = wa[3 * n :]

    ftol = ftol or 1.49012e-8
    xtol = xtol or 1.49012e-8
    gtol = gtol or 0.0
    mode = mode or 1
    factor = factor or 100.0
    nprint = nprint or 0
    diag = ones(n, dtype=x.dtype) if diag is None else diag
    maxfev = maxfev or 200 * (n + 1)

    return _lmder(
        fcn,
        m,
        n,
        x.copy(),
        fvec,
        fjac,
        ldfjac,
        ftol,
        xtol,
        gtol,
        maxfev,
        diag,
        mode,
        factor,
        nprint,
        nfevptr,
        njevptr,
        ipvt,
        qtf,
        wa1,
        wa2,
        wa3,
        wa4,
        udata,
    )



"""_summary_"""

from numba import extending, njit

from ._cminpack import Cminpack
from .utils import _check_dtype


def _chkder(m, n, x, fvec, fjac, ldfjac, xp, fvecp, mode, err):
    raise NotImplementedError


@extending.overload(_chkder)
def _chkder_overload(m, n, x, fvec, fjac, ldfjac, xp, fvecp, mode, err):
    _check_dtype((fvec, fjac, xp, fvecp), x.dtype)
    chkder_external = Cminpack.chkder(x.dtype)

    def impl(m, n, x, fvec, fjac, ldfjac, xp, fvecp, mode, err):
        return chkder_external(
            m,
            n,
            x.ctypes,
            fvec.ctypes,
            fjac.ctypes,
            ldfjac,
            xp.ctypes,
            fvecp.ctypes,
            mode,
            err.ctypes,
        )

    return impl


@njit
def chkder(m, n, x, fvec, fjac, ldfjac, xp, fvecp, mode, err):
    return _chkder(m, n, x, fvec, fjac, ldfjac, xp, fvecp, mode, err)



"""Utility functions for cminpack_numba."""

from __future__ import annotations

import ctypes as ct
import typing
import warnings
from pathlib import Path

from numba.core import cgutils
from numba.extending import intrinsic
from numba.types import CPointer, int32, voidptr

if typing.TYPE_CHECKING:
    from numba.core.typing import Signature
    from numba.types import FunctionType, Type
    from numpy.typing import ArrayLike, DTypeLike, NDArray

ptr_int32 = CPointer(int32)


@intrinsic
def address_as_void_pointer(
    typingctx,
    src: Type,
) -> tuple[Signature, callable]:
    """Void pointer from given memory address.

    Copied from: https://stackoverflow.com/a/61550054/15456681

    Parameters
    ----------
    typingctx :
        typing context
    src : Type
        type of memory address

    Returns
    -------
    tuple[Signature, callable]
        type signature and implementation function for the intrinsic

    """
    sig = voidptr(src)

    def codegen(context, builder, signature, args):
        return builder.inttoptr(args[0], cgutils.voidptr_t)

    return sig, codegen


@intrinsic
def ptr_from_val(typingctx, src: Type) -> tuple[Signature, callable]:
    """Pointer from given value.

    Copied from: https://stackoverflow.com/a/59538114/15456681

    Parameters
    ----------
    typingctx : _type_
        typing context
    src : Type
        type of value

    Returns
    -------
    tuple[Signature, callable]
        type signature and implementation function for the intrinsic

    """

    def impl(context, builder, signature, args):
        return cgutils.alloca_once_value(builder, args[0])

    sig = CPointer(src)(src)
    return sig, impl


@intrinsic
def val_from_ptr(typingctx, src: Type) -> tuple[Signature, callable]:
    """Value from given pointer.

    Copied from: https://stackoverflow.com/a/59538114/15456681

    Parameters
    ----------
    typingctx : _type_
        typing context
    src : Type
        type of pointer

    Returns
    -------
    tuple[Signature, callable]
        type signature and implementation function for the intrinsic

    """

    def impl(context, builder, signature, args):
        return builder.load(args[0])

    sig = src.dtype(src)
    return sig, impl


def _check_dtype(
    args: tuple[ArrayLike, ...],
    dtype: DTypeLike,
    *,
    error: bool = True,
) -> None:
    """Check that all array arguments are of the same dtype.

    Parameters
    ----------
    args : tuple[ArrayLike, ...]
        The array arguments
    dtype : DTypeLike
        The dtype to check against
    error : bool, optional
        Whether to raise an error (warning raised if False), by default True

    Raises
    ------
    ValueError
        If all array arguments are not of the same dtype and error is True

    """
    if not all(i.dtype is dtype for i in args):
        msg = "All array arguments {} be of the same dtype"
        if error:
            raise ValueError(msg.format("must"))
        warnings.warn(msg.format("should"), stacklevel=1)


def check_cfunc(func: FunctionType, *args: NDArray | int) -> ct.c_int:
    """Check a numba cfunc with ctypes.

    Parameters
    ----------
    func : types.FunctionType
        The function to check.
    *args : NDArray | int
        The arguments to pass to the function.

    Returns
    -------
    ct.c_int
        The return value of the function.

    """
    _converter = {
        ct.c_void_p: lambda x: x.ctypes.data,
        ct.c_int: lambda x: x,
        ct.POINTER(ct.c_int): lambda x: x.ctypes.data_as(ct.POINTER(ct.c_int)),
        ct.POINTER(ct.c_double): lambda x: x.ctypes.data_as(ct.POINTER(ct.c_double)),
        ct.POINTER(ct.c_float): lambda x: x.ctypes.data_as(ct.POINTER(ct.c_float)),
    }

    _func = func.ctypes
    _args = [_converter[j](i) for i, j in zip(args, _func.argtypes)]

    return _func(*_args)


def get_extension_path(lib_name: str) -> str:
    """Get the path to the library with the given name in the parent directory.

    Parameters
    ----------
    lib_name : str
        The name of the library to search for.

    Returns
    -------
    str
        The path to the library.

    """
    search_path = Path(__file__).parent
    ext_path = f"**/{lib_name}.*"
    matches = search_path.glob(ext_path)
    try:
        return str(next(matches))
    except StopIteration:
        return None



"""Wrappers of the `dpmpar` and `sdpmpar` functions."""

from numba import extending, njit, types
from numpy import float32, float64, int32

from ._cminpack import Cminpack

# -------------------------------------- dpmpar -------------------------------------- #


def _dpmpar(i):
    raise NotImplementedError


def _sdpmpar(i):
    raise NotImplementedError


@extending.overload(_dpmpar)
def _dpmpar_overload(i):
    dpmpar_external = Cminpack.dpmpar(types.float64)

    def impl(i):
        return float64(dpmpar_external(i))

    return impl


@extending.overload(_sdpmpar)
def _sdpmpar_overload(i):
    sdpmpar_external = Cminpack.dpmpar(types.float32)

    def impl(i):
        return float32(sdpmpar_external(i))

    return impl


@njit
def dpmpar(i: int32) -> float64:
    """Double precision machine parameters.

    Parameters
    ----------
    i : int32
        1: the machine precision
        2: the smallest magnitude
        3: the largest magnitude

    Returns
    -------
    float64
        machine parameter

    """
    return _dpmpar(i)


@njit
def sdpmpar(i: int32) -> float32:
    """Single precision machine parameters.

    Parameters
    ----------
    i : int32
        1: the machine precision
        2: the smallest magnitude
        3: the largest magnitude

    Returns
    -------
    float32
        machine parameter

    """
    return _sdpmpar(i)



"""A numba wrapper for the cminpack library."""

__version__ = "0.1.1"


# TODO(nin17): import functions for implementing bounds

from .chkder import chkder
from .dpmpar import dpmpar, sdpmpar
from .enorm import enorm, enorm_
from .hybrd import hybrd, hybrd1, hybrd1_, hybrd_
from .hybrj import hybrj, hybrj1, hybrj1_, hybrj_
from .lmder import lmder, lmder1, lmder1_, lmder_
from .lmdif import lmdif, lmdif1, lmdif1_, lmdif_
from .lmstr import lmstr, lmstr1, lmstr1_, lmstr_
from .signatures import (
    CminpackSignature,
    hybrd_sig,
    hybrj_sig,
    lmder_sig,
    lmdif_sig,
    lmstr_sig,
    shybrd_sig,
    shybrj_sig,
    slmder_sig,
    slmdif_sig,
    slmstr_sig,
)
from .utils import address_as_void_pointer, check_cfunc, ptr_from_val, val_from_ptr

__all__ = [
    # Utils
    "address_as_void_pointer",
    "check_cfunc",
    "ptr_from_val",
    "val_from_ptr",
    # Signatures
    "hybrd_sig",
    "shybrd_sig",
    "hybrj_sig",
    "shybrj_sig",
    "lmdif_sig",
    "slmdif_sig",
    "lmder_sig",
    "slmder_sig",
    "lmstr_sig",
    "slmstr_sig",
    "CminpackSignature",
    # Functions
    "enorm",
    "enorm_",
    "chkder",
    "dpmpar",
    "sdpmpar",
    "hybrd1",
    "hybrd1_",
    "hybrd",
    "hybrd_",
    "hybrj1",
    "hybrj1_",
    "hybrj",
    "hybrj_",
    "lmdif1",
    "lmdif1_",
    "lmdif",
    "lmdif_",
    "lmder1",
    "lmder1_",
    "lmder",
    "lmder_",
    "lmstr1",
    "lmstr1_",
    "lmstr",
    "lmstr_",
]



"""Signatures for the functions passed to the cminpack functions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from numba import types

if TYPE_CHECKING:
    from numba.core.typing import Signature

__all__ = [
    "hybrd_sig",
    "shybrd_sig",
    "hybrj_sig",
    "shybrj_sig",
    "lmdif_sig",
    "slmdif_sig",
    "lmder_sig",
    "slmder_sig",
    "lmstr_sig",
    "slmstr_sig",
    "CminpackSignature",
]


class CminpackSignature:
    """Signatures for the functions passed to the cminpack functions."""

    @staticmethod
    def hybrd(
        udata_type: types.Type = types.voidptr,
        dtype: types.Float = types.float64,
    ) -> Signature:
        """Signature for the function passed to the `hybrd` & `shybrd` functions.

        Parameters
        ----------
        udata_type : types.Type, optional
            The type of udata, by default types.voidptr
        dtype : types.Float, optional
            The dtype, by default types.float64

        Returns
        -------
        Signature
            The signature of the function

        """
        return types.intc(
            udata_type,  # *udata / *p
            types.intc,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.intc,  # iflag
        )

    @staticmethod
    def hybrj(
        udata_type: types.Type = types.voidptr,
        dtype: types.Float = types.float64,
    ) -> Signature:
        """Signature for the function passed to the `hybrj` & `shybrj` functions.

        Parameters
        ----------
        udata_type : types.Type, optional
            The type of udata, by default types.voidptr
        dtype : types.Float, optional
            The dtype, by default types.float64

        Returns
        -------
        Signature
            The signature of the function

        """
        return types.intc(
            udata_type,  # *udata / *p
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            types.int32,  # iflag
        )

    @staticmethod
    def lmdif(
        udata_type: types.Type = types.voidptr,
        dtype: types.Float = types.float64,
    ) -> Signature:
        """Signature for the function passed to the `lmdif` & `slmdif` functions.

        Parameters
        ----------
        udata_type : types.Type, optional
            The type of udata, by default types.voidptr
        dtype : types.Float, optional
            The dtype, by default types.float64

        Returns
        -------
        Signature
            The signature of the function

        """
        return types.intc(
            udata_type,  # *udata / *p
            types.intc,  # m
            types.intc,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.intc,  # iflag
        )

    @staticmethod
    def lmder(
        udata_type: types.Type = types.voidptr,
        dtype: types.Float = types.float64,
    ) -> Signature:
        """Signature for the function passed to the `lmder` & `slmder` functions.

        Parameters
        ----------
        udata_type : types.Type, optional
            The type of udata, by default types.voidptr
        dtype : types.Float, optional
            The dtype, by default types.float64

        Returns
        -------
        Signature
            The signature of the function

        """
        return types.intc(
            udata_type,  # *udata / *p
            types.intc,  # m
            types.intc,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.intc,  # ldfjac
            types.intc,  # iflag
        )

    @staticmethod
    def lmstr(
        udata_type: types.Type = types.voidptr,
        dtype: types.Float = types.float64,
    ) -> Signature:
        """Signature for the function passed to the `lmstr` & `slmstr` functions.

        Parameters
        ----------
        udata_type : types.Type, optional
            The type of udata, by default types.voidptr
        dtype : types.Float, optional
            The dtype, by default types.float64

        Returns
        -------
        Signature
            The signature of the function

        """
        return types.intc(
            udata_type,  # *udata / *p
            types.intc,  # m
            types.intc,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjrow
            types.intc,  # iflag
        )


# __cminpack_type_fcn_nn__
hybrd_sig = CminpackSignature.hybrd()
"""
(udata: void*, n: int32, x: float64*, fvec: float64*, iflag: int32) -> int32
"""

# __cminpack_type_fcn_nn__
shybrd_sig = CminpackSignature.hybrd(dtype=types.float32)
"""
(udata: void*, n: int32, x: float32*, fvec: float32*, iflag: int32) -> int32
"""

# __cminpack_type_fcnder_nn__
hybrj_sig = CminpackSignature.hybrj()
"""
(udata: void*, n: int32, x: float64*, fvec: float64*, fjac: float64*,
    ldfjac: int32, iflag:int32) -> int32
"""

# __cminpack_type_fcnder_nn__
shybrj_sig = CminpackSignature.hybrj(dtype=types.float32)
"""
(udata: void*, n: int32, x: float32*, fvec: float32*, fjac: float32*,
    ldfjac: int32, iflag:int32) -> int32
"""

# __cminpack_type_fcn_mn__
lmdif_sig = CminpackSignature.lmdif()
"""
(udata: void*, m: int32, n: int32, x: float64*, fvec: float64*, iflag: int32)
    -> int32
"""

# __cminpack_type_fcn_mn_s__
slmdif_sig = CminpackSignature.lmdif(dtype=types.float32)
"""
(udata: void*, m: int32, n: int32, x: float32*, fvec: float32*, iflag: int32)
    -> int32
"""

# __cminpack_type_fcnder_mn__
lmder_sig = CminpackSignature.lmder()
"""
(udata: void*, m: int32, n: int32, x: float64*, fvec: float64*, fjac: float64*,
    ldfjac: int32, iflag: int32) -> int32
"""

# __cminpack_type_fcnder_mn__
slmder_sig = CminpackSignature.lmder(dtype=types.float32)
"""
(udata: void*, m: int32, n: int32, x: float32*, fvec: float32*, fjac: float32*,
    ldfjac: int32, iflag: int32) -> int32
"""

# __cminpack_type_fcnderstr_mn__
lmstr_sig = CminpackSignature.lmstr()
"""
(udata: void*, m: int32, n: int32, x: float64*, fvec: float64*, fjac: float64*,
    iflag: int32) -> int32
"""

# __cminpack_type_fcnderstr_mn__
slmstr_sig = CminpackSignature.lmstr(dtype=types.float32)
"""
(udata: void*, m: int32, n: int32, x: float32*, fvec: float32*, fjac: float32*,
    iflag: int32) -> int32
"""



"""Signatures for the cminpack functions."""

import warnings
from ctypes.util import find_library

from llvmlite import binding
from numba import types

from .utils import get_extension_path

__all__ = [
    "Cminpack",
]


def _apply_prefix(func: str, dtype: types.Float) -> str:
    """Get the cminpack function name with the correct prefix for the given dtype.

    Parameters
    ----------
    func : str
        The function name
    dtype : types.Float
        The dtype

    Returns
    -------
    str
        The cminpack function name with the correct prefix

    """
    _cminpack_prefix = {types.float32: "s", types.float64: ""}
    return f"{_cminpack_prefix[dtype]}{func}"


def _ensure_cminpack(dtype: str = "") -> None:
    """Ensure that the cminpack library is available, in either 64 or 32 bit precision.

    Parameters
    ----------
    dtype : str, optional
        "" for double precision, "s" for single precision, by default ""

    Raises
    ------
    ImportError
        If the cminpack/cminpacks library is not found

    """
    if (
        find_library(f"cminpack{dtype}") is None
        and get_extension_path(f"libcminpack{dtype}") is None
    ):
        msg = f"cminpack{dtype} library not found"
        raise ImportError(msg)


# Load double and single precision libraries
try:
    _cminpack_path = get_extension_path("libcminpack") or find_library("cminpack")
    _ensure_cminpack()
    CMINPACK = True
    binding.load_library_permanently(_cminpack_path)
except ImportError:
    warnings.warn(
        "cminpack not found. Double precision functions unavailable.",
        stacklevel=1,
    )
    CMINPACK = False
try:
    _cminpacks_path = get_extension_path("libcminpacks") or find_library("cminpacks")
    _ensure_cminpack("s")
    CMINPACKS = True
    binding.load_library_permanently(_cminpacks_path)
except ImportError:
    warnings.warn(
        "cminpacks not found. Single precision functions unavailable.",
        stacklevel=1,
    )
    CMINPACKS = False

if not (CMINPACK or CMINPACKS):
    _msg = "cminpack & cminpacks not found."
    raise ImportError(_msg)


class Cminpack:
    """External functions from the cminpack(s) library."""

    @staticmethod
    def chkder(
        dtype: types.Float,
    ) -> types.ExternalFunction:
        """Return the external function for chkder.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for chkder

        """
        sig = types.void(
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            types.CPointer(dtype),  # *xp
            types.CPointer(dtype),  # *fvecp
            types.int32,  # mode
            types.CPointer(dtype),  # *err
        )
        return types.ExternalFunction(_apply_prefix("chkder", dtype), sig)

    @staticmethod
    def dpmpar(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for dpmpar.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for dpmpar

        """
        sig = dtype(types.int32)
        return types.ExternalFunction(_apply_prefix("dpmpar", dtype), sig)

    @staticmethod
    def enorm(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for enorm.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for enorm

        """
        sig = dtype(
            types.int32,  # n
            types.CPointer(dtype),  # *x
        )
        return types.ExternalFunction(_apply_prefix("enorm", dtype), sig)

    @staticmethod
    def hybrd(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for hybrd.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for hybrd

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            dtype,  # xtol
            types.int32,  # maxfev
            types.int32,  # ml
            types.int32,  # mu
            dtype,  # epsfcn
            types.CPointer(dtype),  # *diag
            types.int32,  # mode
            dtype,  # factor
            types.int32,  # nprint
            types.CPointer(types.int32),  # *nfev
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            types.CPointer(dtype),  # *r
            types.int32,  # lr
            types.CPointer(dtype),  # *qtf
            types.CPointer(dtype),  # *wa1
            types.CPointer(dtype),  # *wa2
            types.CPointer(dtype),  # *wa3
            types.CPointer(dtype),  # *wa4
        )
        return types.ExternalFunction(_apply_prefix("hybrd", dtype), sig)

    @staticmethod
    def hybrd1(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for hybrd1.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for hybrd1

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            dtype,  # tol
            types.CPointer(dtype),  # *wa
            types.int32,  # lwa
        )
        return types.ExternalFunction(_apply_prefix("hybrd1", dtype), sig)

    @staticmethod
    def hybrj(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for hybrj.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for hybrj

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            dtype,  # xtol
            types.int32,  # maxfev
            types.CPointer(dtype),  # *diag
            types.int32,  # mode
            dtype,  # factor
            types.int32,  # nprint
            types.CPointer(types.int32),  # *nfev
            types.CPointer(types.int32),  # *njev
            types.CPointer(dtype),  # *r
            types.int32,  # lr
            types.CPointer(dtype),  # *qtf
            types.CPointer(dtype),  # *wa1
            types.CPointer(dtype),  # *wa2
            types.CPointer(dtype),  # *wa3
            types.CPointer(dtype),  # *wa4
        )
        return types.ExternalFunction(_apply_prefix("hybrj", dtype), sig)

    @staticmethod
    def hybrj1(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for hybrj1.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for hybrj1

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            dtype,  # tol
            types.CPointer(dtype),  # *wa
            types.int32,  # lwa
        )
        return types.ExternalFunction(_apply_prefix("hybrj1", dtype), sig)

    @staticmethod
    def lmdif(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for lmdif.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for lmdif

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            dtype,  # ftol
            dtype,  # xtol
            dtype,  # gtol
            types.int32,  # maxfev
            dtype,  # epsfcn
            types.CPointer(dtype),  # *diag
            types.int32,  # mode
            dtype,  # factor
            types.int32,  # nprint
            types.CPointer(types.int32),  # *nfev
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            types.CPointer(types.int32),  # *ipvt
            types.CPointer(dtype),  # *qtf
            types.CPointer(dtype),  # *wa1
            types.CPointer(dtype),  # *wa2
            types.CPointer(dtype),  # *wa3
            types.CPointer(dtype),  # *wa4
        )
        return types.ExternalFunction(_apply_prefix("lmdif", dtype), sig)

    @staticmethod
    def lmdif1(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for lmdif1.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for lmdif1

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            dtype,  # tol
            types.CPointer(types.int32),  # *iwa
            types.CPointer(dtype),  # *wa
            types.int32,  # lwa
        )
        return types.ExternalFunction(_apply_prefix("lmdif1", dtype), sig)

    @staticmethod
    def lmder(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for lmder.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for lmder

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            dtype,  # ftol
            dtype,  # xtol
            dtype,  # gtol
            types.int32,  # maxfev
            types.CPointer(dtype),  # *diag
            types.int32,  # mode
            dtype,  # factor
            types.int32,  # nprint
            types.CPointer(types.int32),  # *nfev
            types.CPointer(types.int32),  # *njev
            types.CPointer(types.int32),  # *ipvt
            types.CPointer(dtype),  # *qtf
            types.CPointer(dtype),  # *wa1
            types.CPointer(dtype),  # *wa2
            types.CPointer(dtype),  # *wa3
            types.CPointer(dtype),  # *wa4
        )
        return types.ExternalFunction(_apply_prefix("lmder", dtype), sig)

    @staticmethod
    def lmder1(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for lmder1.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for lmder1

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            dtype,  # tol
            types.CPointer(types.int32),  # *ipvt
            types.CPointer(dtype),  # *wa
            types.int32,  # lwa
        )
        return types.ExternalFunction(_apply_prefix("lmder1", dtype), sig)

    @staticmethod
    def lmstr(
        dtype: types.Float,
    ) -> types.ExternalFunction:
        """Return the external function for lmstr.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for lmstr

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32,  # ldfjac
            dtype,  # ftol
            dtype,  # xtol
            dtype,  # gtol
            types.int32,  # maxfev
            types.CPointer(dtype),  # *diag
            types.int32,  # mode
            dtype,  # factor
            types.int32,  # nprint
            types.CPointer(types.int32),  # *nfev
            types.CPointer(types.int32),  # *njev
            types.CPointer(types.int32),  # *ipvt
            types.CPointer(dtype),  # *qtf
            types.CPointer(dtype),  # *wa1
            types.CPointer(dtype),  # *wa2
            types.CPointer(dtype),  # *wa3
            types.CPointer(dtype),  # *wa4
        )
        return types.ExternalFunction(_apply_prefix("lmstr", dtype), sig)

    @staticmethod
    def lmstr1(dtype: types.Float) -> types.ExternalFunction:
        """Return the external function for lmstr1.

        Parameters
        ----------
        dtype : types.Float
            The dtype

        Returns
        -------
        types.ExternalFunction
            The external function for lmstr1

        """
        sig = types.int32(
            types.voidptr,  # fcn
            types.voidptr,  # *p / *udata
            types.int32,  # m
            types.int32,  # n
            types.CPointer(dtype),  # *x
            types.CPointer(dtype),  # *fvec
            types.CPointer(dtype),  # *fjac
            types.int32, # ldfjac
            dtype,  # tol
            types.CPointer(types.int32),  # *ipvt
            types.CPointer(dtype),  # *wa
            types.int32,  # lwa
        )
        return types.ExternalFunction(_apply_prefix("lmstr1", dtype), sig)
